require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/artifact_search/artifact_search_query'
require 'uu/os/artifact_search/artifact_search_result_item'

module UU
  module OS
    module ArtifactSearch
      extend self

      #Service path
      PATH = 'ues/core/artifact/UESArtifactSearch'

      # Performs a search for artifacts with specified criteria. The search is performed in the location according to the location_uri, 
      # although it can be widened by using include_system_area or include_preferred_area flag in the search query.
      #
      # Search query contains several search settings - query (what to search,
      # and order settings) and paging settings.
      #
      # **Recommended usage**
      #
      # - if location_uri is an organizational unit, use deepest organizational unit
      #   in the structure, so that the result set is restricted as much as possible.
      #   If the OU (its nesting level) is near the top of the territory, it might
      #   be better for performance to use the territory instead (if possible) - try measuring
      #   times in your case and decide based on that.
      #
      # @param [String, UU::OS::UESURI] location_uri The location to search in. Can be a territory or an organizational unit.
      # @param [ArtifactSearchQuery] search_query DTO containing search criteria. Available query parameters
      # are: {ArtifactSearchResultItem#name name}, {ArtifactSearchResultItem#code code}, {ArtifactSearchResultItem#uri uri},
      # {ArtifactSearchResultItem#organizational_unit_name organizationalUnitName} (cannot use fulltext operator),
      # {ArtifactSearchResultItem#organizational_unit_uri organizationalUnitUri},
      # {ArtifactSearchResultItem#meta_artifact_uri metaArtifactUri},
      # {ArtifactSearchResultItem#meta_artifact_code metaArtifactCode},
      # {ArtifactSearchResultItem#state_type stateType}: {Artifact::ArtifactStateType ArtifactStateType} and
      # {ArtifactSearchResultItem#entity_type_uri entityTypeUri}.
      #
      # See {https://plus4u.net/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      #
      # @return [UU::OS::REST::ResultList<ArtifactSearch::ArtifactSearchQuery, ArtifactSearch::ArtifactSearchResultItem>] The list of found artifacts.
      def query(location_uri, search_query = nil)
        svc = UU::OS::REST::RemoteClient.new(ArtifactSearch)
        dto = UU::OS::ArtifactSearch::ArtifactSearchQuery.new(search_query)
        svc.add_parameter('query', dto.query)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('recursive', dto.recursive)
        svc.add_parameter('includeSystemArea', dto.include_system_area)
        svc.add_parameter('includePreferredArea', dto.include_preferred_area)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('query', location_uri)
          return UU::OS::REST::ResultList.new(ArtifactSearch::ArtifactSearchQuery, ArtifactSearch::ArtifactSearchResultItem, res)
        end
      end
    end
  end
end