require 'uu/os/uesuri'
require 'uu/os/env/version_strategy'
require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/attachment/attachment_attributes'
require 'uu/os/attachment/attachment_check_in'
require 'uu/os/attachment/attachment_check_out'
require 'uu/os/attachment/attachment_create'
require 'uu/os/attachment/attachment_get_attachment_list'
require 'uu/os/attachment/attachment_set_attributes'
require 'uu/os/attachment/attachment_delete_version'
require 'uu/os/attachment/attachment_get_version_attributes'
require 'uu/os/attachment/attachment_get_version_data'
require 'uu/os/attachment/attachment_get_version_list'
require 'uu/os/attachment/attachment_version_attributes'
require 'uu/os/attachment/stream_wrapper'
require 'uu_uds'

module UU
  module OS
    # Attachment service.
    module Attachment
      extend self

      # Service path
      PATH = 'ues/core/attachment/UESAttachment'

      # @private
      ATTACHMENT_BUCKET = 'uu-uds:UNIVERSE:UU.OS/ATTACHMENT/UPLOAD_BUCKET'
      # Creates a new attachment to the specified artifact. At the very least, attachment
      # data and filename have to be specified in this command. The data are streamed.
      #
      # @param [String, UU::OS::UESURI] artifact_uriUESURI of the artifact where the new attachment is created
      # @param [AttachmentCreate] attachment DTO containing attributes of a new attachment
      #
      # @return [UU::OS::UESURI] UESURI of the new attachment
      def create(artifact_uri, attachment = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        payload = Attachment::AttachmentCreate.new(attachment)

        ###############
        ##IMPORTANT!!##
        ###############
        # Attachment.create and check_in are a special case. Unlike other commands that upload bigger data (Sheet.check_in etc.), these
        # don't do it directly by calling the REST api and sending the whole payload to it. Instead we first try to take the data
        # and upload them to UDS binary store. If that succeeds, only the resulting UDS uri is sent as attachment data so the attachment
        # command retrieves the actual data server-side from the UDS service. But should the uds upload fail, we need to send the data
        # the usual way. But some types of stream cant be re-used so we have to make a backup of them so in case of the UDS failure, there
        # is something to send again.
        ###############

        blob_uri = handle_input_data(payload)

        # Blob uri wasnt supplied, we need prepare backup before we proceed with UDS upload
        backup = backup_if_necessary(payload) if blob_uri.nil?

        # Upload data by uds (unless user has supplied his uds uri)
        blob_uri = upload_to_uds(payload.data, backup) unless blob_uri

        UU::OS::QoS::QoSHandler.auto_retry do
          payload = prepare_payload_for_upload(payload,blob_uri,backup)
          begin
            res = svc.post('create', artifact_uri, payload)
            return UU::OS::UESURI.new(res)
          ensure
            silent_cleanup(backup,blob_uri)
          end
        end
      end

      # Returns attributes representing the attachment specified by the attachment_uri attribute. If the
      # attachment is locked by the current user, this command returns attributes of the private version.
      # If the attachment is locked by another user (or if the attachment is unlocked), the command returns
      # attributes of the public version. The command does not change the state of the attachment or its
      # lock in the system.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment
      #
      # @return [AttachmentAttributes] DTO with attachment attributes
      def get_attributes(attachment_uri)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', attachment_uri)
          return Attachment::AttachmentAttributes.new(res)
        end
      end

      # Retrieves a result list of attachments for the artifact specified by the artifact_uri parameter.
      # The resulting attachment list is filtered according to the criteria provided by the criteria
      # attribute. These criteria specify the page to return and name and code filters. The list is sorted
      # (ascending) according to the attachment name and code.
      #
      # @param [String, UU::OS::UESURI] artifact_uri UESURI of the artifact to get the list of attachments from.
      # @param [AttachmentGetAttachmentList] criteria DTO containing paging and filtering criteria.
      #
      # @return [UU::OS::REST::ResultList<Attachment::AttachmentGetAttachmentList, Attachment::AttachmentAttributes>] List of attachments.
      def get_attachment_list(artifact_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        dto = Attachment::AttachmentGetAttachmentList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttachmentList', artifact_uri)
          return UU::OS::REST::ResultList.new(Attachment::AttachmentGetAttachmentList, Attachment::AttachmentAttributes, res)
        end
      end

      # Locks the attachment specified by the attachment_uri attribute. If the operation is successful
      # or if the attachment is already locked by the current user, attachment stream handler or nil is returned
      # depending on the options set in the attribute options. If the attachment cannot be locked because the
      # attachment is locked by another user, the command will fail.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment
      # @param [AttachmentCheckOut] options DTO with options containing a flag whether to return the stream with attachment data
      #
      # @return [UU::OS::REST::BinaryValue] Data content of the attachment or nil depending on the options set in options attribute
      def check_out(attachment_uri, options = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        dto = Attachment::AttachmentCheckOut.new(options)
        payload = dto.to_json
        UU::OS::QoS::QoSHandler.auto_retry do
          if dto.return_stream_handler == TRUE
            return svc.post_binary('checkOut', attachment_uri, payload)
          else
            svc.post('checkOut', attachment_uri, payload)
            return nil
          end
        end
      end

      # Updates the attached file in the system and unlocks it by default. The command succeeds if the
      # attachment is unlocked or if it is locked by the current user. Otherwise, the command fails. At the very least,
      # attachment data and the filename have to be specified in this command. The data are streamed.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment
      # @param [AttachmentCheckIn] attachment DTO containing attributes of the checked-in attachment
      def check_in(attachment_uri, attachment = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        payload = Attachment::AttachmentCheckIn.new(attachment)

        ###############
        ##IMPORTANT!!##
        ###############
        # Attachment.create and check_in are a special case. Unlike other commands that upload bigger data (Sheet.check_in etc.), these
        # don't do it directly by calling the REST api and sending the whole payload to it. Instead we first try to take the data
        # and upload them to UDS binary store. If that succeeds, only the resulting UDS uri is sent as attachment data so the attachment
        # command retrieves the actual data server-side from the UDS service. But should the uds upload fail, we need to send the data
        # the usual way. But some types of stream cant be re-used so we have to make a backup of them so in case of the UDS failure, there
        # is something to send again.
        ###############

        blob_uri = handle_input_data(payload)

        # Blob uri wasnt supplied, we need prepare backup before we proceed with UDS upload
        backup = backup_if_necessary(payload) if blob_uri.nil?

        # Upload data by uds (unless user has supplied his uds uri)
        blob_uri = upload_to_uds(payload.data, backup) unless blob_uri

        UU::OS::QoS::QoSHandler.auto_retry do
          payload = prepare_payload_for_upload(payload,blob_uri,backup)
          begin
            svc.post('checkIn', attachment_uri, payload)
          ensure
            silent_cleanup(backup,blob_uri)
          end
        end
      end

      # Returns stream handler containing data content of the attachment. If the attachment is locked by the
      # same user, command returns data of the private version. If the attachment is locked by another user (or if
      # the attachment is unlocked), command returns data of the public version. Command does not change the state
      # of the attachment or its lock in the system.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment
      #
      # @return [UU::OS::REST::BinaryValue] DTO with attachment data content
      def get_attachment_data(attachment_uri)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getAttachmentData', attachment_uri)
        end
      end

      # Deletes the specified attachment.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment which is about to be deleted.
      def delete(attachment_uri)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', attachment_uri)
        end
      end

      # Updates basic attributes of an attachment.
      #
      # @param [String, UU::OS::UESURI] attachment_uriUESURI of the attachment whose attributes are to be changed.
      # @param [AttachmentSetAttributes] attachment DTO containing the attachment attributes.
      #
      # @return [] UESURI of the attachment
      def set_attributes(attachment_uri, attachment = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        payload = UU::OS::Attachment::AttachmentSetAttributes.new(attachment).to_json
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', attachment_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Deletes an attachment version from the system.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment whose version is going to be deleted.
      # @param [AttachmentDeleteVersion] version Object containing ID of the version which is to be deleted.
      # @option version [String] :version_id Not null. Id of the attachment version, which is going to be deleted.
      #
      # @example
      #   # Delete attachment version (uri needs to be replaced with one of an existing attachment for the example to work)
      #   UU::OS::Attachment.delete_version('ues:TERRITORY:ARTIFACT:ATTACH', :version_id => "123456789")
      def delete_version(attachment_uri, version = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        payload = UU::OS::Attachment::AttachmentDeleteVersion.new(version).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('deleteVersion', attachment_uri, payload)
        end
      end

      # Retrieves the data of the specified attachment version.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment whose version is going to be retrieved.
      # @param [AttachmentGetVersionData] version Specific options.
      # @option version [String] :version_id Not null. Id of the attachment version, which data are going to be retrieved.
      #
      # @return [UU::OS::REST::BinaryValue] Binary data containing the data of the specified attachment version.
      #
      # @example
      #   # Get version of the attachment 'ues:TER:ART:ATTACH' identified by its OID = 456232122
      #   UU::OS::Attachment.get_version_data('ues:TER:ART:ATTACH', :version_id => "456232122")
      def get_version_data(attachment_uri, version)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        dto = UU::OS::Attachment::AttachmentGetVersionData.new(version)
        svc.add_parameter('versionId', dto.version_id)
        UU::OS::QoS::QoSHandler.auto_retry do
          result = svc.get_binary('getVersionData', attachment_uri)
          return result
        end
      end

      # The uuCommand returns attributes of given attachment version.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment whose version's attributes are going to be retrieved.
      # @param [AttachmentGetVersionAttributes] version Object containing ID of the searched version.
      # @option version [String] :version_id Not null. ID of the attachment version, which attributes are going to be retrieved.
      #
      # @return [AttachmentVersionAttributes] Object with attributes of the attachment version.
      #
      # @example
      #   # Get attachment version attributes (uri needs to be replaced with one of an existing attachment for the example to work)
      #   attributes = UU::OS::Attachment.get_version_attributes('ues:TERRITORY:ARTIFACT:ATTACH', :version_id => "123456789")

      def get_version_attributes(attachment_uri, version = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        dto = UU::OS::Attachment::AttachmentGetVersionAttributes.new(version)
        svc.add_parameter('versionId', dto.version_id)

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getVersionAttributes', attachment_uri)
          return Attachment::AttachmentVersionAttributes.new(res)
        end
      end

      # The uuCommand returns a list of the specified attachment's versions.
      #
      # @param [String, UU::OS::UESURI] attachment_uri UESURI of the attachment whose versions are going to be listed.
      # @param [AttachmentGetVersionList] criteria Criteria for selecting versions.
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {uuQuery - Guideline}[https://plus4u.net/ues/sesm?SessFree=ues%253AUNI-BT%255B210795%255D%253AUES.DOC%252FCMD_V1%252FUESQUERY%255B44191574519623267%255D%253A]). It is possible to filter the result list by name, code, label, creationTime, creationStamp and version.
      #
      # @return [Array] An array with the command result. Contains instances of {UU::OS::Attachment::AttachmentVersionAttributes}.
      #
      # @example
      #   # Get attachment versions ordered by version in descending order (uri needs to be replaced with one of an existing attachment for the example to work)
      #   list = UU::OS::Attachment.get_version_list('ues:TERRITORY:ARTIFACT:ATTACH', :query => "ORDER BY version DESC")
      def get_version_list(attachment_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Attachment)
        dto = UU::OS::Attachment::AttachmentGetVersionList.new(criteria)          
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getVersionList', attachment_uri)
          return UU::OS::REST::ResultList.new(Attachment::AttachmentGetVersionList, Attachment::AttachmentVersionAttributes, res)
        end
      end

      private

      def check_for_uri_in_string_data(str)
        if str.start_with?('uu-uds:') and str.size < 4000
          begin
            return UU::OS::UESURI.new(str)
          rescue Exception => e
            return nil
          end
        end
      end

      def upload_to_uds(value,backup = nil)

        unless backup.nil?
          value.data = backup
        end

        return UU::UDS::Bucket.create_blob(ATTACHMENT_BUCKET,value) rescue nil
      end

      def convert_to_filename(name)
        # return default name of file if is not given
        return 'data' if name.nil? || name.strip.empty?
        # strip and replace white spaces and vertical tabulator
        name.strip.gsub(/[\s\u000B]/, '_')
      end

      def create_stream_with_file_name(bin_val)
        StreamWrapper.new(bin_val.data, "/#{bin_val.name}")
      end

      def create_file_with_file_name(bin_val)
        # Reset stream before each call in order to send all data
        file = File.new(bin_val.data.path,'rb')

        path = bin_val.name.nil? || bin_val.name.strip.empty? ? file.path : "/#{bin_val.name}"
        StreamWrapper.new(file, path)
      end

      # Backup data stream in case of uds upload failure
      def backup_if_necessary(payload)
        backup = nil

        # Only special cases of stream need backup (not File, that can be rewinded/reopened)
        if (payload.data.data.kind_of?StringIO)
          backup = payload.data.data.clone
        elsif (payload.data.data.kind_of?IO) && (!payload.data.data.kind_of?File)
          backup = Tempfile.new(payload.data.name)
          IO.copy_stream(payload.data.data,backup)
        end

        return backup
      end

      # Blob uri is set, so we can assume upload to UDS succeeded or user provided his own uri
      def prepare_payload_for_upload(payload,blob_uri,backup)
        if blob_uri
          # Ensure we're not sending real data again (should be in UDS now)
          payload.data = nil
          # Ensure uploading as multipart/form (hashes are sent that way)
          payload = payload.to_hash
          payload[:dataUri] = blob_uri
        else
          # If we got here, lets assume the UDS upload failed and we need to upload directly to attachment command

          # Use stream from backup if available
          unless backup.nil?
            payload.data.data = backup
          end

          # All data types are in BinaryValue
          bin_val = payload.data

          # Reset stream before each call in order to send all data
          payload.data = create_file_with_file_name(bin_val) if bin_val.data.kind_of?File
          # add method for file name recognition by rest client TODO add support to remote client directly
          payload.data = create_stream_with_file_name(bin_val) unless bin_val.data.kind_of?File
          payload = payload.to_hash
          payload[:data] = payload.delete(:data)
          begin
            payload[:data].rewind if payload[:data].respond_to? :rewind
          rescue IOError => e #remoteclient v1
            if "closed stream" == e.message
              payload[:data].open
            else
              raise e
            end
          end
        end

        return payload
      end

      def silent_cleanup(backup,blob_uri)
        unless backup.nil?
          backup.close unless backup.closed? rescue nil
          backup.unlink if backup.respond_to?(:unlink) rescue nil
        end
        if blob_uri
          UU::UDS::Bucket.delete_blob(blob_uri) rescue nil
        end
      end

      # This method is meant to accept all kinds of input data types for user comfort and prepare them
      # for uploading as BinaryValue also this method detectes if user isnt trying to upload actual data
      # but just supplied own uds uri, in that case the method returns the uri, otherwise it doesnt return
      # anything and just manipulates payload
      def handle_input_data(payload)
        blob_uri=nil

        # first we check all types of data and do conversions/wrapping
        if payload.data.kind_of?Hash
          # Hash should be interpreted as BinaryValue
          payload.data = UU::OS::REST::BinaryValue.new(payload.data)
        elsif payload.data.kind_of?String
          # User supplied a string, it could contain UDS uri so we need to check for that
          uri_in_str = check_for_uri_in_string_data(payload.data)
          if uri_in_str
            blob_uri = uri_in_str unless uri_in_str.nil?
          else
            # Otherwise treat it as a regular string and generate a filename for it
            payload.data = UU::OS::REST::BinaryValue.new(payload.data)
            payload.data.name = "#{convert_to_filename(payload.name)}.txt"
          end
        elsif payload.data.kind_of?UESURI
          # User supplied an UESURI, this type is accepted only when its an uri of UDS blob (will be checked later)
          blob_uri = payload.data
        elsif payload.data.kind_of?File
          # A regular file, just wrap it with binary value
          payload.data = UU::OS::REST::BinaryValue.new(payload.data)
        elsif payload.data.kind_of?IO
          # A stream. Unlike File it doesnt have name so generate one (*.bin)
          payload.data = UU::OS::REST::BinaryValue.new(payload.data)
          payload.data.name = "#{convert_to_filename(payload.name)}.bin"
        elsif payload.data.kind_of?StringIO
          # A user-made StringIO, doesnt have filename, so generate one (*.txt)
          payload.data = UU::OS::REST::BinaryValue.new(payload.data)
          payload.data.name = "#{convert_to_filename(payload.name)}.txt"
        elsif payload.data.kind_of?UU::OS::REST::BinaryValue
          # User supplied a BinaryValue, thats ok but lets check if the name isnt empty and if it is
          # generate one (*.txt/*.bin)
          if (payload.data.data.kind_of?String or payload.data.data.kind_of?StringIO)
            extension = 'txt'
          else
            extension = 'bin'
          end
          payload.data.name = "#{convert_to_filename(payload.name)}.#{extension}" if payload.data.name.nil? || payload.data.name.strip.empty?
        else
          raise ArgumentError.new("Data can be only instance of String, StringIO, IO, File, or BinaryValue, but was #{payload.data.class}.")
        end
        return blob_uri
      end

    end

  end
end
