require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/folder/folder_create'
require 'uu/os/folder/folder_attributes'
require 'uu/os/folder/folder_entry_attributes'
require 'uu/os/folder/folder_export'
require 'uu/os/folder/folder_get_entry_list'
require 'uu/os/folder/folder_get_export_data'
require 'uu/os/folder/folder_set_attributes'

module UU
  module OS
    module Folder
      extend self

      # Service path
      PATH = 'ues/core/container/UESFolder'
      # Creates a folder of class Folder. Command creates new folder in specified container.
      # At the very least meta artifact must be specified in {UU::OS::Artifact::ArtifactCreate artifactCreate}
      # Object. Competent role for artifact will be selected as the most suitable according
      # to specified container and meta artifact (executive/authorized role with connected
      # interface), or can be also specified in {ArtifactCreate} Object. 
      #
      # @param [String, UU::OS::UESURI] location_uri UESURI of the folder/org. unit/meta model where folder will be created.
      # @param [UU::OS::Folder::FolderCreate] folder Specific options.
      # @option folder [String, UU::OS::UESURI] :meta_artifact_uri UESURI of the meta artifact. Only meta artifact with template class of Folder (not descendants of Folder) can be used. If meta artifact is not allowed to be created in container specified in first parameter of command UESFolder.create(UESURI, UESFolderCreate), create command will fail. 
      # @option folder [String] :code Code of the new folder. When null is set, the code will be generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails. 
      # @option folder [String] :name Name of the folder. When null is set, meta artifact's tName will be used if set, otherwise the create command will fail. 
      # @option folder [String] :description Description of the folder. When null is set, meta artifact's tDescription will be used if set, otherwise no description will be set on the created folder. 
      # @option folder [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role. When null is set, competent role will be taken from authorization. Executive role for creating artifact into container specified in first parameter of command UESFolder.create(UESURI, UESFolderCreate), will be used. If no role is authorized to create artifact into specified container, create command will fail. If selected role has not connected role interface with selected meta artifact, create command will fail. 
      # @option folder [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. When null is set, security level will be taken from meta artifact tSecurityLevel. If meta artifact tSecurityLevel is higher than user's security clearance, the create command will fail. If security level is higher than users security clearance, create command will fail. 
      # @option folder [String, UU::OS::UESURI] :icon_uri URI of the folder's icon. When null is set, the folder uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option folder [Boolean] :dry_run Specifies whether this is only a test run to verify that the object can be created. The default value is false.
      # @option folder [Boolean] :component_content Defines whether the primarily displayed content of the created Folder is component content. Default value is false.
      # 
      # @return [UU::OS::UESURI] UESURI of the created folder
      #       
      # @example
      #   # Create new folder.
      #   UU::OS::Folder.create('ues:TERRITORY:ORGANIZATIONAL_UNIT', :code => 'NEW_FOLDER', :name => 'New Folder')
      #  
      #   # Verifying that new folder can be created.
      #   UU::OS::Folder.create('ues:TERRITORY:ORGANIZATIONAL_UNIT', :code => 'NEW_FOLDER', :name => 'New Folder', :dryRun => true)      
      def create(location_uri, folder = nil)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        payload = UU::OS::Folder::FolderCreate.new(folder).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', location_uri, payload)
          # res is nil if dry_run was set to true. In that case, return nil (do nothing)
          if !(res.nil? || res.empty?)
            return UU::OS::UESURI.new(res)
          end
        end
      end

      # Reads folder specified by the <code>folderUri</code> attribute and returns
      # its attributes. 
      #
      # @param [String, UU::OS::UESURI] folder_uri UESURI of the folder 
      #
      # @return [FolderAttributes] Object with folder attributes
      # 
      # @example
      #   # Get attributes of specified folder.
      #   getAttributes = UU::OS::Folder.get_attributes('ues:TERRITORY:FOLDER')
      # 
      def get_attributes(folder_uri)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', folder_uri)
          return Folder::FolderAttributes.new(res)
        end
      end

      # This command returns a list of entries (artifacts or shortcuts) enlisted in the
      # specified folder (entity type of main entity of the command could be a Folder, Organizational
      # Unit, Territory or Meta Model Dictionary). Command does not return all entries,
      # only a sublist. The list of returned entries could be filtered by the name, code
      # or type of an entry. The list is sorted by the name (and codes in case are equal) of
      # a entry by default, or it can be sorted arbitrarily by any of entry attributes usable
      # for the filtering or these attributes combinations. 
      #
      # @param [String, UU::OS::UESURI] folder_uri UESURI of the folder from where entries (artifacts or shortcuts) are listed. It could
      # be a Folder, Organizational Unit, Territory or Meta Model Dictionary. 
      # @param [FolderGetEntryList] criteria Object containing criteria for listing shortcuts. Available parameters are: {FolderEntryAttributes#name name}, {FolderEntryAttributes#code code} and {FolderEntryAttributes#type type} : {FolderEntryType}. See {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY UESQuery Documentation}.
      # @option criteria [String] :query The <b>UES query</b> string. It has to conform the ues query syntax (see documentation for more info). Filtering and order of the result list can be specified there. When null is set, no filtering is applied and entries are ordered by name and code as the second criteria.
      # @option criteria [Boolean] :recursive Flag for recursive listing. It means that if the listed folder contains inner folders and the recursive flag is true, entries of the all inner folders are in result too. When the recursive flag is null or false, only entries are returned exactly from the listed folder.
      #
      # @return [UU::OS::REST::ResultList<Folder::FolderGetEntryList, Folder::FolderEntryAttributes>] List of Objects representing folder entries (artifacts or shortcuts).
      # 
      # @example
      #   # Get entries of folder in descending order and even with entries of inner folders - flag recursive is set to true.
      #   getEntryList = UU::OS::Folder.get_entry_list('ues:TERRITORY:FOLDER', :query=>'ORDER BY name DESC', :recursive=>true)
      #       
      def get_entry_list(folder_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        dto = Folder::FolderGetEntryList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        svc.add_parameter('recursive', dto.recursive)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getEntryList', folder_uri)
          return UU::OS::REST::ResultList.new(Folder::FolderGetEntryList, Folder::FolderEntryAttributes, res)
        end
      end

      # This command deletes the specified folder from the system. The folder is deleted
      # even if it contains a locked sheet or attachment. The folder cannot be deleted when:
      #   the folder is not empty.
      #   the folder's workflow is not created 
      #   a property contains a reference to this folder 
      #   a time sheet exists with a reference to this folder
      #
      # @param [String, UU::OS::UESURI] folder_uri UESURI of the folder which is about to be deleted.
      # 
      # @example
      #   # Delete folder with specified uri.
      #   UU::OS::Folder.delete('ues:TERRITORY:FOLDER')
      #
      def delete(folder_uri)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', folder_uri)
        end
      end

      # Command for setting attributes of a folder. The command can't modify references
      # to objects like location, competent role and so on. 
      #
      # @param [String, UU::OS::UESURI] folderUri UESURI of the folder whose attributes are to be changed. 
      # @param [FolderSetAttributes] folder Object containing new attributes the folder. 
      # @option folder [String] :code Code of the folder. When nil is set, the code is not modified.
      # @option folder [String] :name Name of the folder. When nil is set, the name is not modified.
      # @option folder [String] :description Description of the folder. When nil is set, the description is not modified.
      # @option folder [String, UU::OS::Artifact::SecurityLevel] :security_level Security level of the folder. When nil is set, the security level is not modified.
      # @option folder [String, UU::OS::UESURI] :icon_uri URI of the folder's icon. When nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the folder icon is removed and the folder uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option folder [Boolean] :component_content Defines whether the primarily displayed content of the created Folder is component content. Default value is false.
      #
      # @return [UU::OS::UESURI] UESURI of the folder whose attributes are to be changed.
      # 
      # @example
      #   # Set folder basic attributes.
      #   UU::OS::Folder.set_attributes('ues:TERRITORY:FOLDER', :code=>'NEW_CODE', :name=>'Folder name', :description=>'Some sample description.')
      #  
      def set_attributes(folder_uri, folder = nil)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        payload = UU::OS::Folder::FolderSetAttributes.new(folder).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', folder_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Exports a folder to the XML file, which is saved to the export storage. The export is an asynchronous process. <br>
      # It is possible to wait for the end of the process. More information in {UU::OS::Env::Process} and in {UU::OS::REST::Future}.
      #
      # @param [String, UU::OS::UESURI] folder_uri UESURI of the folder to export.
      #
      # @return [UU::OS::UESURI] UESURI of the export process.
      #
      # @example
      #   # Start exporting a folder.
      #   process_uri = UU::OS::Folder.export('ues:TERRITORY:FOLDER')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(proces).get(:timeout=>600)
      def export(folder_uri)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        payload = UU::OS::Folder::FolderExport.new(nil).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('export', folder_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a XML file generated by the export command {Folder#export Folder Export}.
      #
      # @param [String, UU::OS::UESURI] folder_uri UESURI of the folder we want to get the exported XML file from.
      # @param [UU::OS::Folder::FolderGetExportData] folder Specific options.
      # @option folder [String] :requested_mime_type MIME type of the returned value. All possible MIME types can be found as constants in {UU::OS::Artifact::ExportMimeType} class. ZIP is returned by default.
      #
      # @return [UU::OS::REST::BinaryValue] Data file of an exported folder (XML or ZIP containing XML).
      #
      # @example
      #   # Start exporting a folder
      #   process_uri = UU::OS::Folder.export('ues:TERRITORY:FOLDER')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::Folder.get_export_data(export_uri, :requested_mime_type => 'application/xml')
      def get_export_data(folder_uri, folder = nil)
        svc = UU::OS::REST::RemoteClient.new(Folder)
        dto = UU::OS::Folder::FolderGetExportData.new(folder)
        if dto.requested_mime_type
          svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        end
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getExportData', folder_uri)
        end
      end

    end
  end

end
