require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/meta_visual_use_case/meta_visual_use_case_copy'
require 'uu/os/meta_visual_use_case/meta_visual_use_case_attributes'
require 'uu/os/meta_visual_use_case/meta_visual_use_case_set_attributes'



module UU
  module OS
    module MetaVisualUseCase
      extend self

      # Service path
      PATH = '/uu/os/MetaVisualUseCase'

      # Copy a meta visual use case. The command creates a new meta visual use case
      # according to specified source meta visual use case in a specified location.
      # At the very least location meta visual use case and name must be specified
      # in {MetaVisualUseCaseCopy} DTO. Competent role for the new meta visual use case
      # copy will be selected as the most suitable according to specified container and
      # role(executive/authorized role with connected interface), or can be also specified
      # in {MetaVisualUseCaseCopy} DTO.
      # 
      # @param [String, UU::OS::UESURI] source_meta_artifact_uri UESURI of the meta visual use case which will be copied
      # @param [MetaVisualUseCaseCopy] meta_visual_use_case DTO containing attributes of the new meta visual use case copy
      # @option meta_visual_use_case [String, UU::OS::UESURI] :location_uri UESURI of the meta model dictionary where to the meta visual use case will be copied. If the rights does not permit to copy a meta visual use case from the meta visual use case in this location, the copy command fails. The value must not be nil.
      # @option meta_visual_use_case [String] :name Name of the new meta visual use case. If nil is set, default name from meta artifact is used if set, otherwise the create command fails.
      # @option meta_visual_use_case [String] :code Code of the new meta visual use case. If nil is set, the code is generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails.
      # @option meta_visual_use_case [String] :description Description of the new meta visual use case. If nil is set, default description from meta artifact is used if set, otherwise no description is set on the new meta visual use case.
      # @option meta_visual_use_case [String, UU::OS::UESURI] :competent_role_uri UESURI of the competent role. If nil is set, the executive role from authorization is used. If no role is authorized to create the role into specified container, the create command fails. If the selected role doesn't have connected the role interface with the selected meta artifact, the create command fails. If the selected role doesn't belong to the same organizational unit as a new role or if it doesn't belong to one of superior organizational units, the create command fails.
      # @option meta_visual_use_case [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. If nil is set, security level is taken from the source meta visual use case security level. If the source meta visual use case security level is higher than the user's security clearance, the copy command fails.
      # @option meta_visual_use_case [String, UU::OS::UESURI] :icon_uri UESURI of the meta visual use case's icon. If nil is set, the meta visual use case uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_visual_use_case [Boolean] :copy_sheets Specifies whether to copy content from the source meta visual use case to the new copy. Default value is false.
      # @option meta_visual_use_case [Boolean] :copy_lifecycle Specifies whether to copy life cycle from the source meta visual use case to the new copy. Default value is false.
      # @option meta_visual_use_case [Boolean] :copy_attachments Specifies whether to copy attachments from the source meta visual use case to the new copy. Default value is false.
      # @option meta_visual_use_case [Boolean] :copy_sheet_templates Specifies whether to copy sheet templates from the source meta visual use case to the new copy. Default value is false.
      # @option meta_visual_use_case [String] :default_name Specifies default name of the created visual use case according to new meta visual use case copy. If nil is set, the default name is taken from the source meta visual use case.
      # @option meta_visual_use_case [String] :default_code Specifies default code of the created visual use case according to new meta visual use case copy. If nil is set, the default code is taken from the source meta visual use case.
      # @option meta_visual_use_case [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies default security level of the created visual use case according to new meta visual use case copy. If nil is set, the default security level is taken from the source meta visual use case.
      # @option meta_visual_use_case [String, UU::OS::UESURI] :default_icon_uri Specifies default icon of the created visual use case according to new meta visual use case copy. If nil is set, the default icon is taken from the source meta visual use case. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_visual_use_case [Boolean] :component_content Defines whether the primarily displayed content of the uuEE is component content. Default value is false.
      # @option meta_visual_use_case [Boolean] :default_component_content Defines whether the primarily displayed content of a visual use case created according to the new meta visual use case copy is component content. When nil is set, the default component content setting is copied from the source meta artifact.
      #
      # @return [UU::OS::UESURI] UESURI of the new meta visual use case
      #
      # @example
      #   # Copy meta visual use case
      #   UU::OS::MetaVisualUseCase.copy('ues:TERRITORY:META_ARTIFACT',
      #        :location_uri=>'ues:TERRITORY:LOCATION',
      #        :name=>'Copy name',
      #        :code=>'COPY_CODE',
      #        :description=>'Some description string',
      #        :competent_role_uri=>'ues:TERRITORY:COMPETENT_ROLE',
      #        :security_level=>UU::OS::Artifact::SecurityLevel::TOP_SECRET,
      #        :copy_sheets=>false,
      #        :copy_lifecycle=>false,
      #        :copy_attachments=>true,
      #        :copy_sheet_templates=>false,
      #        :default_name=>'Default name',
      #        :default_code=>'DEFAULT_CODE',
      #        :default_security_level=>UU::OS::Artifact::SecurityLevel::CONFIDENTIAL
      #   )
      def copy(source_meta_artifact_uri, meta_visual_use_case = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaVisualUseCase)
        payload = UU::OS::MetaVisualUseCase::MetaVisualUseCaseCopy.new(meta_visual_use_case).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('copy', source_meta_artifact_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns attributes representing the meta visual use case specified by the meta_visual_use_case_uri
      # parameter.
      # 
      # @param [String, UU::OS::UESURI] meta_visual_use_case_uri UESURI of the MVUC for which we'll fetch its attributes
      #
      # @return [MetaVisualUseCaseAttributes] DTO containing the MVUC attributes
      #
      # @example
      #   # Get meta visual use case attributes.
      #   attributes = UU::OS::MetaVisualUseCase.get_attributes('ues:TERRITORY:META_VISUAL_USE_CASE')
      def get_attributes(meta_visual_use_case_uri)
        svc = UU::OS::REST::RemoteClient.new(MetaVisualUseCase)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', meta_visual_use_case_uri)
          return MetaVisualUseCase::MetaVisualUseCaseAttributes.new(res)
        end
      end

      # Sets attributes on specified meta visual use case according to meta_visual_use_case_uri.
      # 
      # @param [String, UU::OS::UESURI] meta_visual_use_case_uri UESURI of the MVUC for which we'll set new attributes
      # @param [MetaVisualUseCaseSetAttributes] meta_visual_use_case DTO containing new attributes
      # @option meta_visual_use_case [String] :name The new name of the meta visual use case. If nil is set, the name is not modified.
      # @option meta_visual_use_case [String] :code The new code of the meta visual use case. If nil is set, the code is not modified. If blank string is set, the code is generated automatically using the system sequence. The code must be unique in a given territory, otherwise the set_attributes command fails.
      # @option meta_visual_use_case [String] :description The new description of the meta visual use case. If nil is set, the description is not modified.
      # @option meta_visual_use_case [String, UU::OS::Artifact::SecurityLevel] :security_level The new security level of the meta visual use case. If nil is set, the security level is not modified. If user is setting the meta visual use case template security level higher than the user's security clearance, the set_attributes command fails.
      # @option meta_visual_use_case [String, UU::OS::UESURI] :icon_uri URI of the new meta visual use case's icon. If nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the meta visual use case icon is removed and the meta visual use case uses icon from its meta visual use case. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_visual_use_case [String] :default_name Specifies new default name of the created visual use case according to meta visual use case (specified in source_meta_artifact_uri). If nil is set, the default name is not modified.
      # @option meta_visual_use_case [String] :default_code Specifies new default code of the created visual use case according to meta visual use case (specified in source_meta_artifact_uri). If nil is set, the default code is not modified.
      # @option meta_visual_use_case [String, UU::OS::Artifact::SecurityLevel] :default_security_level Specifies new default security level of the created visual use case according to meta visual use case (specified in source_meta_artifact_uri). If nil is set, the default security is not modified.
      # @option meta_visual_use_case [String, UU::OS::UESURI] :default_icon_uri Specifies new default icon of the created visual use case according to meta visual use case (specified in meta_artifact_uri). If nil is set, the default icon is not modified. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option meta_visual_use_case [Boolean] :component_content Defines whether the primarily displayed content of the uuEE is component content. Default value is false.
      # @option meta_visual_use_case [Boolean] :default_component_content Defines whether the primarily displayed content of a visual use case created according to the meta visual use case (specified in source_meta_artifact_uri) is component content. When nil is set, the default component content setting is not modified.
      #
      # @return [UU::OS::UESURI] UESURI of the MVUC after attributes change
      #
      # @example
      #   # Set attributes of meta artifact.
      #   UU::OS::MetaArtifact.set_attributes('ues:TERRITORY:META_ARTIFACT',
      #           :name=>'Meta art name',
      #           :code=>'CODE',
      #           :description=>'Text description without sense.',
      #           :security_level=>UU::OS::Artifact::SecurityLevel::CONFIDENTIAL,
      #           :default_name=>'Mar default name',
      #           :default_code=>'DEFAULT_CODE',
      #           :default_security_level=>UU::OS::Artifact::SecurityLevel::CONFIDENTIAL
      #           )
      def set_attributes(meta_visual_use_case_uri, meta_visual_use_case = nil)
        svc = UU::OS::REST::RemoteClient.new(MetaVisualUseCase)
        payload = UU::OS::MetaVisualUseCase::MetaVisualUseCaseSetAttributes.new(meta_visual_use_case).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', meta_visual_use_case_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

    end
  end
end
